//
//  SupaConvex.m
//
//  Created by Konstantin Kiselyov on 9/24/12.
//  Copyright (c) 2012 Konstantin Kiselyov. All rights reserved.
//

#import "SupaConvex.h"

#define eps 1e-3

@interface SupaConvex()
@property (assign) BOOL isValidConvex;
@property (strong, nonatomic) NSArray* convexPoints;

- (void)invalidateConvex;
- (void)updateConvexPoints;
@end

BOOL more(float a, float b) {
    return (a - b) >= eps;
}

BOOL eq(float a, float b) {
    return fabs(a - b) < eps;
}

BOOL moreeq(float a, float b) {
    return more(a, b) || eq(a, b);
}

BOOL less(float a, float b) {
    return !moreeq(a, b);
}

BOOL lesseq(float a, float b) {
    return !more(a, b);
}

double cross_product(CGPoint v1, CGPoint v2) {
    return v1.x*v2.y - v2.x*v1.y;
}

static BOOL cw (CGPoint a, CGPoint b, CGPoint c) {
	return less(a.x*(b.y-c.y)+b.x*(c.y-a.y)+c.x*(a.y-b.y), 0);
}

static BOOL ccw (CGPoint a, CGPoint b, CGPoint c) {
	return more(a.x*(b.y-c.y)+b.x*(c.y-a.y)+c.x*(a.y-b.y), 0);
}

@implementation SupaConvex
@synthesize convexPoints, allPoints, isValidConvex, allRects;

- (id)init {
    self = [super init];
    if ( !self ) {
        return nil;
    }
    self.allPoints = [NSArray array];
    self.convexPoints = [NSArray array];
    self.isValidConvex = YES;
    self.allRects = [NSArray array];
    return self;
}

- (void)invalidateConvex {
    self.isValidConvex = NO;
}

- (void)addPoint:(CGPoint)point {
    NSMutableArray* allPointsMutable = [[NSMutableArray alloc] initWithArray: self.allPoints];
    [allPointsMutable addObject: [NSValue valueWithCGPoint: point]];
    self.allPoints = [NSArray arrayWithArray: allPointsMutable];
    [self invalidateConvex];
}

- (void)setAllPoints:(NSArray *)allPoints_ {
    allPoints = allPoints_;
    [self invalidateConvex];
}

- (void)addRect:(CGRect)rect toLeft:(NSMutableArray *)leftPoints {
    [leftPoints addObject: [NSValue valueWithCGPoint: CGPointMake(rect.origin.x, rect.origin.y)]];
    [leftPoints addObject: [NSValue valueWithCGPoint: CGPointMake(rect.origin.x, rect.origin.y + rect.size.height)]];
}

- (void)addRect:(CGRect)rect toRight:(NSMutableArray *)rightPoints {
    [rightPoints addObject: [NSValue valueWithCGPoint: CGPointMake(rect.origin.x + rect.size.width, rect.origin.y)]];
    [rightPoints addObject: [NSValue valueWithCGPoint: CGPointMake(rect.origin.x + rect.size.width, rect.origin.y + rect.size.height)]];
}

- (void)updateConvexPoints {
    NSMutableArray* convexMutable = [[NSMutableArray alloc] initWithCapacity: 0];
    if ( !self.allRects ) {
        self.convexPoints = [NSArray array];
    }
    
    NSArray* sortedRects = [self.allRects sortedArrayUsingComparator: ^NSComparisonResult(id obj1, id obj2) {
        CGRect r1 = [obj1 CGRectValue];
        CGRect r2 = [obj2 CGRectValue];
        if ( r1.origin.y < r2.origin.y ) {
            return NSOrderedAscending;
        }
        return NSOrderedDescending;
    }];
    
//    NSLog(@"Sorted selection rects:\n%@\n", sortedRects);
    
    NSMutableArray* leftPoints = [[NSMutableArray alloc] initWithCapacity: 0];
    NSMutableArray* rightPoints = [[NSMutableArray alloc] initWithCapacity: 0];
    
    CGRect lastRect = [[sortedRects objectAtIndex: 0] CGRectValue];
    [self addRect: lastRect toLeft: leftPoints];
    [self addRect: lastRect toRight: rightPoints];
    
    for ( int i = 1; i < sortedRects.count; ++i ) {
        CGFloat curY = lastRect.origin.y + lastRect.size.height;
        
        CGRect curRect = [[sortedRects objectAtIndex: i] CGRectValue];
        CGFloat spaceHeight = curRect.origin.y - curY;
        CGFloat firstSpaceHeight = floorf(spaceHeight/2);
        CGFloat secondSpaceHeight = spaceHeight - firstSpaceHeight;

        CGRect firstSpaceRect = lastRect, secondSpaceRect = curRect;

        firstSpaceRect.origin.y = lastRect.origin.y + lastRect.size.height;
        firstSpaceRect.size.height = firstSpaceHeight;
        
        secondSpaceRect.origin.y = curRect.origin.y - secondSpaceHeight;
        secondSpaceRect.size.height = secondSpaceHeight;

//        [self addRect: firstSpaceRect toLeft: leftPoints];
//        [self addRect: firstSpaceRect toRight: rightPoints];
//
//        [self addRect: secondSpaceRect toLeft: leftPoints];
//        [self addRect: secondSpaceRect toRight: rightPoints];
        
        [self addRect: curRect toLeft: leftPoints];
        [self addRect: curRect toRight: rightPoints];
    }

    NSArray* reversedRightPoints = [[rightPoints reverseObjectEnumerator] allObjects];
    [convexMutable addObjectsFromArray: leftPoints];
    [convexMutable addObjectsFromArray: reversedRightPoints];
    
    for ( int i = 0; i < convexMutable.count; ++i ) {
        CGPoint cur = [[convexMutable objectAtIndex: i] CGPointValue];
        CGPoint next = [[convexMutable objectAtIndex: (i+1)%(convexMutable.count)] CGPointValue];
        
        // Work around
        if ( more(fabs(cur.x - next.x), fabs(cur.y - next.y)) ) {
            cur.y = floorf((cur.y + next.y)/2);
            next.y = cur.y;
        }
        else {
            cur.x = floorf((cur.x + next.x)/2);
            next.x = cur.x;
        }
        
        [convexMutable replaceObjectAtIndex: i withObject: [NSValue valueWithCGPoint: cur]];
        [convexMutable replaceObjectAtIndex: (i+1)%(convexMutable.count) withObject: [NSValue valueWithCGPoint: next]];
    }
    
    self.convexPoints = [NSArray arrayWithArray: convexMutable];

    // Validate convex
    self.isValidConvex = YES;
}

- (void)addRect:(CGRect)rect {
    NSMutableArray* allRectsMutable = [[NSMutableArray alloc] initWithArray: self.allRects];
    [allRectsMutable addObject: [NSValue valueWithCGRect: rect]];
    
    self.allRects = [NSArray arrayWithArray: allRectsMutable];
    [self invalidateConvex];
}

- (NSArray *)convexPoints {
    if ( !self.isValidConvex ) {
        [self updateConvexPoints];
    }
    return convexPoints;
}

@end
