//
//  SupaLabel.m
//
//  Created by Konstantin Kiselyov on 9/23/12.
//  Copyright (c) 2012 Konstantin Kiselyov. All rights reserved.
//

#import "SupaLabel.h"
#import "SupaConvex.h"

#define kFontAttrKey    @"font"
#define kColorAttrKey   @"color"
#define kRangeAttrKey   @"rage"

typedef void (^SupaLabelBeforeTextDrawing_t)(id arg);
typedef void (^SupaLabelAfterTextDrawing_t)(id arg);

@interface SupaLabel()
@property (strong, nonatomic) NSArray* markedRanges;
@property (strong, nonatomic) NSValue* selectionRange;
@property (strong, nonatomic) NSAttributedString* markersAttributedText;

@property (strong, nonatomic) NSDictionary* fontAttr;
@property (strong, nonatomic) NSDictionary* colorAttr;

- (void)initialize;
- (NSAttributedString *)setupAdditionalAttributesForString:(NSString *)attrText;
- (CFDictionaryRef)createClippingPathWithPath:(CGPathRef)pathRef;

- (CTFrameRef)prepareFrame:(CTFrameRef)frame initialString:(NSAttributedString *)initialString inContext:(CGContextRef)context inRect:(CGRect)rect verticalAlignment:(SupaLabelVerticalAlignment_t)verticalAlignment withDots:(BOOL)dotsNeeded;

- (void)drawFrame:(CTFrameRef)frame context:(CGContextRef)context rect:(CGRect)rect beforeText:(SupaLabelBeforeTextDrawing_t)beforeDrawing afterText:(SupaLabelAfterTextDrawing_t)afterText;

// Helpers
- (NSArray *)selectionRectsWithFrame:(CTFrameRef)frame;
- (CGPathRef)pathFromPointsArray:(NSArray *)points;
- (void)clearCFArray:(NSArray *)array;
@end

@implementation SupaLabel
@synthesize attributedText, markedRanges, selectionRange, markersAttributedText, verticalAlignment, columnPadding, columnsNumber, lineSpacing, lineBreakModeKind, textAlignmentHor, clippingPaths, selectionCornerRadius, fontAttr, colorAttr, fillSelectionCallback, padding, selectionColor;

- (void)initialize {
    self.columnsNumber = 1;
    self.columnPadding = 20.0f;
    self.markedRanges = [NSArray array];
    self.markersAttributedText = nil;
    self.lineSpacing = 0;
    self.textAlignmentHor = kCTTextAlignmentLeft;
    self.lineBreakModeKind = kCTLineBreakByWordWrapping;
    self.verticalAlignment = SL_VERTICAL_ALIGNMENT_TOP;
    self.clippingPaths = [NSArray array];
    self.selectionRange = nil;
    self.selectionCornerRadius = 0;
    self.fontAttr = nil;
    self.colorAttr = nil;
    self.fillSelectionCallback = nil;
    self.padding = UIEdgeInsetsMake(0, 0, 0, 0);
    self.selectionColor = [UIColor whiteColor];
}

- (id)initWithFrame:(CGRect)frame
{
    self = [super initWithFrame:frame];
    if (self) {
        [self initialize];
    }
    return self;
}

- (id)initWithCoder:(NSCoder *)aDecoder {
    self = [super initWithCoder: aDecoder];
    if (self) {
        [self initialize];
    }
    return self;
}

- (id)init {
    self = [super init];
    if (self) {
        [self initialize];
    }
    return self;
}

- (void)clearCFArray:(NSArray *)array {
    
}

- (CFDictionaryRef)createClippingPathWithPath:(CGPathRef)pathRef {
    NSDictionary* ret = [NSDictionary dictionaryWithObject: (__bridge id)(pathRef) forKey: (NSString *)kCTFramePathClippingPathAttributeName];
    return (__bridge CFDictionaryRef)(ret);
}

- (NSAttributedString *)setupAdditionalAttributesForString:(NSString *)attrText {
    if ( !attrText || attrText.length == 0 ) {
        return nil;
    }
    // Setup paragraph attributes
    /*
     
     { kCTParagraphStyleSpecifierLineSpacing, sizeof(CGFloat), ...},
     { kCTParagraphStyleSpecifierAlignment, sizeof(CTTextAlignment), ...},
     { kCTParagraphStyleSpecifierFirstLineHeadIndent, sizeof(CGFloat), ...},
     { kCTParagraphStyleSpecifierHeadIndent, sizeof(CGFloat), ...},
     { kCTParagraphStyleSpecifierTabStops, sizeof(CFArrayRef), ...},
     { kCTParagraphStyleSpecifierParagraphSpacing, sizeof(CGFloat), ...},
     { kCTParagraphStyleSpecifierParagraphSpacingBefore, sizeof(CGFloat), ...}
     
     */
    
    CGFloat lineSpacing_ = self.lineSpacing;
    CTTextAlignment horAlignment_ = self.textAlignmentHor;
    CTLineBreakMode lineBreakMode_ = self.lineBreakModeKind;
    
    CTParagraphStyleSetting parSet[] = {
        {kCTParagraphStyleSpecifierAlignment, sizeof(horAlignment_), &horAlignment_},
        {kCTParagraphStyleSpecifierLineSpacing, sizeof(lineSpacing_), &lineSpacing_},
        {kCTParagraphStyleSpecifierLineBreakMode, sizeof(lineBreakMode_), &lineBreakMode_}
    };
    
    NSInteger optsNum = sizeof(parSet) / sizeof(CTParagraphStyleSetting);
    CTParagraphStyleRef parStyle = CTParagraphStyleCreate( parSet, optsNum);

    
    NSAttributedString* ret = nil;
    if (attrText.length > 0) {
        NSMutableAttributedString* attrStrMutable = [[NSMutableAttributedString alloc] initWithString: attrText attributes: [NSDictionary dictionaryWithObject: (__bridge id)parStyle forKey: (NSString *)kCTParagraphStyleAttributeName]];
        if ( self.fontAttr ) {
            UIFont* font = [self.fontAttr objectForKey: kFontAttrKey];
            NSRange range = [[self.fontAttr objectForKey: kRangeAttrKey] rangeValue];
            CTFontRef fontRef = CTFontCreateWithName( (__bridge CFStringRef)((NSString *)font.fontName), font.pointSize, NULL );
            
            if ( range.location + range.length - 1 < attrStrMutable.length ) {
                [attrStrMutable addAttribute: (NSString *)kCTFontAttributeName value: (__bridge id)fontRef range: range];
            }
        }
        
        if ( self.colorAttr ) {
            UIColor* color = [self.colorAttr objectForKey: kColorAttrKey];
            NSRange range = [[self.colorAttr objectForKey: kRangeAttrKey] rangeValue];

            if ( range.location + range.length - 1 < attrStrMutable.length ) {
                [attrStrMutable addAttribute: (NSString *)kCTForegroundColorAttributeName value: (__bridge id)color.CGColor range: range];
            }
        }
        ret = [[NSAttributedString alloc] initWithAttributedString: attrStrMutable];
    }
    return ret;
}

- (CTFrameRef)prepareFrame:(CTFrameRef)frame initialString:(NSAttributedString *)initialString inContext:(CGContextRef)context inRect:(CGRect)rect verticalAlignment:(SupaLabelVerticalAlignment_t)verticalAlignment withDots:(BOOL)dotsNeeded {
    
    CFRange wholeRange = CTFrameGetStringRange(frame);
    CFRange visibleRange = CTFrameGetVisibleStringRange(frame);
    
    CGPathRef framePath = CTFrameGetPath(frame);
    CFDictionaryRef frameOptions = CTFrameGetFrameAttributes(frame);
    
    NSMutableAttributedString* newAttrTextMutable = [[NSMutableAttributedString alloc] initWithAttributedString: initialString];
    
    // Setup specific line-break mode attributes for
    // last visible line of text
    
    if ( (visibleRange.length < wholeRange.length) && dotsNeeded ) {
        NSArray* lines = (__bridge NSArray *)(CTFrameGetLines(frame));
        CTLineRef lastVisibleLine = (__bridge CTLineRef)([lines lastObject]);
        CFRange lastVisibleLineRange = CTLineGetStringRange(lastVisibleLine);
        
        CTLineBreakMode lastVisibleLineBreakMode = kCTLineBreakByTruncatingTail;
        CTParagraphStyleSetting lastLineSetting[1] = {kCTParagraphStyleSpecifierLineBreakMode, sizeof(lastVisibleLineBreakMode), &lastVisibleLineBreakMode};
        CTParagraphStyleRef lastLineStyle = CTParagraphStyleCreate(lastLineSetting, sizeof(lastLineSetting)/sizeof(CTParagraphStyleSetting));
        
        [newAttrTextMutable addAttribute: (NSString *)kCTParagraphStyleAttributeName value: (__bridge id)lastLineStyle range: (NSRange){lastVisibleLineRange.location, lastVisibleLineRange.length}];
        
        CFRelease(lastLineStyle);
    }

    // Create new frame for changed attributes of text
    CTFramesetterRef frameSetter = CTFramesetterCreateWithAttributedString((__bridge CFAttributedStringRef)(newAttrTextMutable));
    frame = CTFramesetterCreateFrame(frameSetter, wholeRange, framePath, frameOptions);
    return frame;
}

- (void)drawFrame:(CTFrameRef)frame context:(CGContextRef)context rect:(CGRect)rect beforeText:(SupaLabelBeforeTextDrawing_t)beforeDrawing afterText:(SupaLabelAfterTextDrawing_t)afterDrawing {
    
    CGContextSaveGState(context);
    
    CGFloat shiftY = 0;
    NSArray* lines = (__bridge NSArray *)(CTFrameGetLines(frame));
    
    // Try calculate y-shift for setting up
    // vertical alignment
    
    if ( lines.count > 0 ) {
        CGPoint lineOrigins[lines.count];
        CTFrameGetLineOrigins(frame, CFRangeMake(0, lines.count), lineOrigins);
        
        CTLineRef firstLine = (__bridge CTLineRef)([lines objectAtIndex: 0]);
        CTLineRef lastLine = (__bridge CTLineRef)([lines lastObject]);
        
        CGFloat firstLineAscent, firstLineDescent, firstLineLeading;
        CGFloat lastLineAscent, lastLineDescent, lastLineLeading;
        
        CTLineGetTypographicBounds(firstLine, &firstLineAscent, &firstLineDescent, &firstLineLeading);
        CTLineGetTypographicBounds(lastLine, &lastLineAscent, &lastLineDescent, &lastLineLeading);
        
        CGFloat topY = lineOrigins[0].y + firstLineAscent - firstLineDescent + 1;
        CGFloat bottomY = lineOrigins[lines.count - 1].y - lastLineDescent + 1;
        
        switch (self.verticalAlignment) {
            case SL_VERTICAL_ALIGNMENT_TOP:
                break;
            case SL_VERTICAL_ALIGNMENT_CENTER:
                shiftY = -floorf(topY - (rect.size.height + fabs(topY - bottomY))/2);
                break;
            case SL_VERTICAL_ALIGNMENT_BOTTOM:
                shiftY = -bottomY;
                break;
            default:
                break;
        }
    }
    // Set transform for vertical alignment
    CGContextTranslateCTM(context, 0, shiftY);
    
    // Make drawing under text
    if ( beforeDrawing ) {
        beforeDrawing(nil);
    }
    
    // Draw text
    CTFrameDraw(frame, context);
    
    // Make drawing above text
    if ( afterDrawing ) {
        afterDrawing(nil);
    }
    CGContextRestoreGState(context);
}

- (CGPathRef)pathFromPointsArray:(NSArray *)points {
    if ( !points || points.count < 2 ) {
        return NULL;
    }
    NSMutableArray* morePoints = [[NSMutableArray alloc] initWithCapacity: 0];
    for ( int i = 0; i < points.count; ++i ) {
        CGPoint curPoint = [[points objectAtIndex: i] CGPointValue];
        CGPoint nextPoint = [[points objectAtIndex: (i + 1)%(points.count)] CGPointValue];
        
        CGPoint newCurPoint = curPoint, newNextPoint = nextPoint;
        if ( more(fabs(curPoint.y - nextPoint.y), fabs(curPoint.x - nextPoint.x)) ) {
            CGFloat sign = ((curPoint.y - nextPoint.y) > 0) ? -1 : 1;
            newCurPoint.y = curPoint.y + sign * self.selectionCornerRadius;
            newNextPoint.y = nextPoint.y - sign * self.selectionCornerRadius;
        }
        else {
            CGFloat sign = ((curPoint.x - nextPoint.x) > 0) ? -1 : 1;
            newCurPoint.x = curPoint.x + sign * self.selectionCornerRadius;
            newNextPoint.x = nextPoint.x - sign * self.selectionCornerRadius;
        }
        [morePoints addObject: [NSValue valueWithCGPoint: newCurPoint]];
        if ( self.selectionCornerRadius > 0 ) {
            [morePoints addObject: [NSValue valueWithCGPoint: newNextPoint]];
        }
    }
    
    CGMutablePathRef mutablePath = CGPathCreateMutable();
    
    CGPoint lastPoint = [[morePoints objectAtIndex: 0] CGPointValue];
    CGPathMoveToPoint(mutablePath, NULL, lastPoint.x, lastPoint.y);
    
    for ( int i = 0; i < morePoints.count; ++i ) {
        CGPoint curPoint = [[morePoints objectAtIndex: i] CGPointValue];
        CGPoint nextPoint = [[morePoints objectAtIndex: (i+1)%(morePoints.count)] CGPointValue];

        if ( more(fabs(curPoint.x - nextPoint.x), 0) && more(fabs(curPoint.y - nextPoint.y), 0) ) {
            CGPoint nextNextPoint = [[morePoints objectAtIndex: (i+2)%(morePoints.count)] CGPointValue];
            CGPoint firstTangentPoint, secondTangentPoint;
            
            if ( more(fabs(nextNextPoint.x - nextPoint.x), fabs(nextNextPoint.y - nextPoint.y)) ) {
                firstTangentPoint = CGPointMake(curPoint.x, nextPoint.y);
                secondTangentPoint = nextPoint;
            }
            else {
                firstTangentPoint = CGPointMake(nextPoint.x, curPoint.y);
                secondTangentPoint = nextPoint;
            }
            CGPathAddArcToPoint(mutablePath, NULL, firstTangentPoint.x, firstTangentPoint.y, secondTangentPoint.x, secondTangentPoint.y, self.selectionCornerRadius);
        }
        else {
            CGPathAddLineToPoint(mutablePath, NULL, nextPoint.x, nextPoint.y);
        }
    }
    return mutablePath;
}

- (NSArray *)selectionRectsWithFrame:(CTFrameRef)frame {
    if ( !self.selectionRange ) {
        return [NSArray array];
    }
    NSMutableArray* rects = [[NSMutableArray alloc] initWithCapacity: 0];
    
    NSRange selectionRange_ = [self.selectionRange rangeValue];
    NSArray* lines = (__bridge NSArray *)(CTFrameGetLines(frame));
    
    CGPoint lineOrigins[lines.count];
    CTFrameGetLineOrigins(frame, CFRangeMake(0, lines.count), lineOrigins);
    
    for ( NSInteger i = 0; i < lines.count; ++i ) {
        CTLineRef line = (__bridge CTLineRef)([lines objectAtIndex: i]);
        CFRange lineRange = CTLineGetStringRange(line);
        
        CGFloat lineAscent, lineDescent, lineLeading;
        CGFloat lineWidth = CTLineGetTypographicBounds(line, &lineAscent, &lineDescent, &lineLeading);
        
        NSInteger startIndex = MAX(lineRange.location, selectionRange_.location);
        NSInteger finishIndex = MIN(lineRange.location + lineRange.length - 1, selectionRange_.location + selectionRange_.length - 1);
        
        if (startIndex <= finishIndex) {
            CGFloat selLeftOffset = CTLineGetOffsetForStringIndex(line, startIndex, NULL);
            CGFloat selRightOffset;
            if ( finishIndex < lineRange.location + lineRange.length - 1 ) {
                selRightOffset = CTLineGetOffsetForStringIndex(line, finishIndex + 1, NULL);
            }
            else {
                selRightOffset = lineWidth;
            }
            
            CGFloat lineHeight = lineAscent + lineDescent + lineLeading;
            CGFloat lineOriginY = lineOrigins[i].y + lineAscent;
            
            CGFloat selTopOffset = lineOriginY - lineHeight;
            CGFloat selWidth = selRightOffset - selLeftOffset + 1;
            CGFloat selHeight = lineHeight;
            
            CGRect selRect = CGRectMake(selLeftOffset, selTopOffset, selWidth, selHeight);
            [rects addObject: [NSValue valueWithCGRect: selRect]];
        }
    }
    return [NSArray arrayWithArray: rects];
}

- (void)addClippingRect:(CGRect)clippingRect padding:(UIEdgeInsets)padding_ {
    NSMutableArray* pathsMutable = [[NSMutableArray alloc] initWithArray: self.clippingPaths];

    CGAffineTransform transform;
    transform = CGAffineTransformMakeTranslation(self.padding.left, self.padding.top);
    transform = CGAffineTransformMakeTranslation(0, self.bounds.size.height);
    transform = CGAffineTransformScale(transform, 1, -1);
    
    CGMutablePathRef clippingPathMutable = CGPathCreateMutable();
    CGPathAddRect(clippingPathMutable, &transform, CGRectMake(clippingRect.origin.x - padding_.left, clippingRect.origin.y - padding_.top, clippingRect.size.width + padding_.left + padding_.right, clippingRect.size.height + padding_.top + padding_.bottom));
    
    [pathsMutable addObject: (__bridge id)(clippingPathMutable)];
    self.clippingPaths = [NSArray arrayWithArray: pathsMutable];
    
    CFRelease(clippingPathMutable);
}

- (void)removeAllClippingPaths {
    self.clippingPaths = [NSArray array];
}

- (void)drawRect:(CGRect)rect {
    rect.size.width -= (self.padding.left + self.padding.right);
    rect.size.height -= (self.padding.top + self.padding.bottom);
    
    if ( !self.attributedText ) {
        self.attributedText = [self setupAdditionalAttributesForString: self.text];
    }
    NSAttributedString* attrText = self.attributedText;
    if ( !attrText ) {
        attrText = [[NSAttributedString alloc] initWithString: self.text];
    }
    
    CGContextRef context = UIGraphicsGetCurrentContext();
    
    CGContextSetTextMatrix( context, CGAffineTransformIdentity );
    CGContextTranslateCTM( context, self.padding.left, self.padding.top );
    CGContextTranslateCTM( context, 0, rect.size.height );
    CGContextScaleCTM( context, 1, -1 );
    
    CTFramesetterRef frameSetter = CTFramesetterCreateWithAttributedString( (__bridge CFAttributedStringRef)attrText );
    
    CGFloat overallTextWidth = rect.size.width;
    CGFloat overallTextHeight = rect.size.height;
    
    CGFloat colWidth = (overallTextWidth - self.columnPadding*(self.columnsNumber-1))/self.columnsNumber;
    CGFloat colHeight = overallTextHeight;
    
    CGMutablePathRef columnRectPaths[self.columnsNumber];
    NSMutableArray* colRects = [[NSMutableArray alloc] initWithCapacity: self.columnsNumber];
    
    for ( NSInteger columnIdx = 0; columnIdx < MAX(self.columnsNumber, 1); ++columnIdx ) {
        CGMutablePathRef rectPath = CGPathCreateMutable();;
        CGFloat originX = (colWidth + self.columnPadding)*columnIdx;
        CGFloat originY = 0;
        
        CGRect colRect = CGRectMake(originX, originY, colWidth, colHeight);
        
        CGPathAddRect(rectPath, NULL, colRect);
        columnRectPaths[columnIdx] = rectPath;
        
        [colRects addObject: [NSValue valueWithCGRect: colRect]];
    }
    
    NSDictionary* frameOptionsDict = nil;
    NSMutableArray* clippingPathDicts = [[NSMutableArray alloc] initWithCapacity: 0];
    
    if ( self.clippingPaths ) {
        for ( int i = 0; i < self.clippingPaths.count; ++i ) {
            CGPathRef path = (__bridge CGPathRef)([self.clippingPaths objectAtIndex: i]);
            [clippingPathDicts addObject: (__bridge NSDictionary *)[self createClippingPathWithPath: path]];
        }
        frameOptionsDict = [NSDictionary dictionaryWithObject: clippingPathDicts forKey: (NSString *)kCTFrameClippingPathsAttributeName];
    }
    
    NSMutableArray* allSelectionRects = [[NSMutableArray alloc] initWithCapacity: 0];
    CTFrameRef colFrames[self.columnsNumber];
    NSInteger nextIndex = 0;
    
    // Prepare column frames
    for ( NSInteger columnIdx = 0; columnIdx < MAX(self.columnsNumber, 1); ++columnIdx ) {
        CFRange colTextRange = CFRangeMake(nextIndex, attrText.length);
        
        CTFrameRef tempFrame = CTFramesetterCreateFrame(frameSetter, colTextRange, columnRectPaths[columnIdx], (__bridge CFDictionaryRef)frameOptionsDict);
        
        CTFrameRef preparedColFrame = [self prepareFrame: tempFrame initialString: attrText inContext: context inRect: [[colRects objectAtIndex: columnIdx] CGRectValue] verticalAlignment: self.verticalAlignment withDots: YES];
        
        colFrames[columnIdx] = preparedColFrame;
        
        NSArray* selectionRects = [self selectionRectsWithFrame: preparedColFrame];
        [allSelectionRects addObjectsFromArray: selectionRects];
        
        CFRange visibleRange = CTFrameGetVisibleStringRange(preparedColFrame);
        nextIndex = visibleRange.location + visibleRange.length;
        
        [self drawFrame: preparedColFrame context: context rect: [[colRects objectAtIndex: columnIdx] CGRectValue] beforeText:^(id arg) {
            // Under text
            if ( selectionRects && selectionRects.count > 0 ) {
                // If only 2 selection rects (1 at the end, and 1 at the beginning)
                
                NSMutableArray* selectionPaths = [[NSMutableArray alloc] initWithCapacity: 0];
                if ( selectionRects.count < 2 ) {
                    for ( int i = 0; i < selectionRects.count; ++i ) {
                        CGRect selRect = [[selectionRects objectAtIndex: i] CGRectValue];
                        NSMutableArray* selPoints = [[NSMutableArray alloc] initWithCapacity: 0];
                        [selPoints addObject: [NSValue valueWithCGPoint: CGPointMake(selRect.origin.x, selRect.origin.y)]];
                        [selPoints addObject: [NSValue valueWithCGPoint: CGPointMake(selRect.origin.x, selRect.origin.y + selRect.size.height)]];
                        [selPoints addObject: [NSValue valueWithCGPoint: CGPointMake(selRect.origin.x + selRect.size.width, selRect.origin.y + selRect.size.height)]];
                        [selPoints addObject: [NSValue valueWithCGPoint: CGPointMake(selRect.origin.x + selRect.size.width, selRect.origin.y)]];
                        
                        CGPathRef selectionPath = [self pathFromPointsArray: selPoints];
                        [selectionPaths addObject: (__bridge id)selectionPath];
                        CGContextAddPath(context, selectionPath);
                    }
                }
                else {
                    SupaConvex* selectionConvex = [[SupaConvex alloc] init];
                    for ( int i = 0; i < selectionRects.count; ++i ) {
                        CGRect selRect = [[selectionRects objectAtIndex: i] CGRectValue];
                        [selectionConvex addRect: selRect];
                    }
                    NSArray* convexPoints = selectionConvex.convexPoints;
                    
                    CGPathRef selectionConvexPath = [self pathFromPointsArray: convexPoints];
                    CGContextAddPath(context, selectionConvexPath);
                    
                    [selectionPaths addObject: (__bridge id)(selectionConvexPath)];
                    CFRelease(selectionConvexPath);
                }
                
                if ( self.fillSelectionCallback ) {
                    CGContextSaveGState(context);
                    
                    self.fillSelectionCallback(context, selectionPaths, rect);
                    
                    CGContextRestoreGState(context);
                }
                else {
                    const float color[] = {1, 0, 0, 0.5};
                    CGContextSetFillColor(context, color);
                    CGContextFillPath(context);
                }
            }
         
        } afterText: ^(id arg) {
            // Above text
            
        }];
        
        CFRelease(tempFrame);
    }
    
    for ( NSInteger columnIdx = 0; columnIdx < self.columnsNumber; ++columnIdx ) {
        CFRelease(colFrames[columnIdx]);
        CFRelease(columnRectPaths[columnIdx]);
    }
    CFRelease( frameSetter );
}

- (void)selectRange:(NSRange)selectionRange_ {
    self.selectionRange = [NSValue valueWithRange: selectionRange_];
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)deselectAll {
    self.selectionRange = nil;
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)markTextWithPattern:(NSString *)pattern {
//    if ( !self.attrText ) {
//        return;
//    }
//    self.selectionsAttributedText = self.attributedText;
//    NSMutableArray* ranges = [[NSMutableArray alloc] initWithCapacity: 0];
//    __autoreleasing NSError* error = nil;
//    
//    NSRegularExpression* regexp = [NSRegularExpression regularExpressionWithPattern: pattern options: NSRegularExpressionCaseInsensitive error: &error];
//    
//    NSMutableAttributedString* attrTextMutable = [[NSMutableAttributedString alloc] initWithAttributedString: self.selectionsAttributedText];
//    
//    [regexp enumerateMatchesInString: [self.selectionsAttributedText string] options: 0 range: (NSRange){0,self.selectionsAttributedText.length} usingBlock:^(NSTextCheckingResult *result, NSMatchingFlags flags, BOOL *stop) {
//        // On match
//        [ranges addObject: [NSValue valueWithRange: [result range]]];
//        [attrTextMutable addAttribute: (__bridge id)kCTForegroundColorAttributeName value: (__bridge id)[UIColor redColor].CGColor range: [result range]];
//    }];
//    
//    self.selectionsAttributedText = [[NSAttributedString alloc] initWithAttributedString: attrTextMutable];
//    self.selectionRanges = [NSArray arrayWithArray: ranges];
//    [self setNeedsDisplay];
//    [self setNeedsLayout];
}

- (void)updateWithAttributedString:(NSAttributedString *)attrText_ {
    self.attributedText = attrText_;
}

- (void)setAttributedText:(NSAttributedString *)attributedText_ {
    attributedText = attributedText_;
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)setText:(NSString *)text {
    [super setText: text];
    self.attributedText = [self setupAdditionalAttributesForString: text];
    
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)setLineSpacing:(CGFloat)lineSpacing_ {
    lineSpacing = lineSpacing_;
    self.attributedText = [self setupAdditionalAttributesForString: self.text];
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)setVerticalAlignment:(SupaLabelVerticalAlignment_t)verticalAlignment_ {
    verticalAlignment = verticalAlignment_;
    self.attributedText = [self setupAdditionalAttributesForString: self.text];
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)setColumnsNumber:(NSInteger)columnsNumber_ {
    columnsNumber = columnsNumber_;
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)setColumnPadding:(CGFloat)columnPadding_ {
    columnPadding = columnPadding_;
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)setLineBreakModeKind:(CTLineBreakMode)lineBreakMode_ {
    lineBreakModeKind = lineBreakMode_;
    self.attributedText = [self setupAdditionalAttributesForString: self.text];
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)setTextAlignmentHor:(CTTextAlignment)textAlignment_ {
    textAlignmentHor = textAlignment_;
    self.attributedText = [self setupAdditionalAttributesForString: self.text];
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)setFont:(UIFont *)font_ forRange:(NSRange)fontRange {
    self.fontAttr = [NSDictionary dictionaryWithObjectsAndKeys: font_, kFontAttrKey, [NSValue valueWithRange: fontRange], kRangeAttrKey, nil];
    self.attributedText = [self setupAdditionalAttributesForString: self.text];
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

- (void)setForegroundColor:(UIColor *)color_ forRange:(NSRange)colorRange {
    self.colorAttr = [NSDictionary dictionaryWithObjectsAndKeys: color_, kColorAttrKey, [NSValue valueWithRange: colorRange], kRangeAttrKey, nil];
    self.attributedText = [self setupAdditionalAttributesForString: self.text];
    [self setNeedsDisplay];
    [self setNeedsLayout];
}

@end
